#!/usr/bin/env python3

# Copyright (C) 2021-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# type: ignore

import argparse
import csv
import logging
import os
from pathlib import Path
import shutil
import subprocess

from swh.export.exporters.edges import GraphEdgesExporter
from swh.export.exporters.orc import ORCExporter
from swh.export.fullnames import process_fullnames
from swh.export.journalprocessor import _add_person
from swh.graph.example_dataset import DATASET
from swh.graph.webgraph import compress
from swh.model.model import Release, Revision


def main():
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(description="Generate a test dataset")
    parser.add_argument(
        "--compress",
        action="store_true",
        default=False,
        help="Also compress the dataset",
    )
    parser.add_argument(
        "--profile", default="release", help="rust profile to use for compression"
    )
    parser.add_argument("output", help="output directory", nargs="?", default=".")
    parser.add_argument(
        "sensitive_output", help="sensitive output directory", nargs="?", default=None
    )
    args = parser.parse_args()

    exporters = {"edges": GraphEdgesExporter, "orc": ORCExporter}
    object_types = [
        "origin",
        "origin_visit",
        "origin_visit_status",
        "snapshot",
        "release",
        "revision",
        "directory",
        "content",
        "skipped_content",
    ]
    config = {"test_unique_file_id": "all"}
    output_path = Path(args.output)
    sensitive_output_path = (
        Path(args.sensitive_output) if args.sensitive_output is not None else None
    )

    if sensitive_output_path is not None:
        if (sensitive_output_path / "orc/person").exists():
            shutil.rmtree(sensitive_output_path / "orc/person")

        sensitive_output_path.mkdir(parents=True, exist_ok=True)

        tmp_sensitive_dir = sensitive_output_path / "tmp"
        tmp_sensitive_dir.mkdir(parents=True, exist_ok=True)

        tmp_dup_dir = tmp_sensitive_dir / "duplicated"
        tmp_dup_dir.mkdir(parents=True, exist_ok=True)

        tmp_dedup_dir = tmp_sensitive_dir / "deduplicated"
        tmp_dedup_dir.mkdir(parents=True, exist_ok=True)

        (sensitive_output_path / "orc/person").mkdir(parents=True, exist_ok=True)

    for name, exporter in exporters.items():
        if (output_path / name).exists():
            shutil.rmtree(output_path / name)
        with exporter(config, object_types, output_path / name) as e:
            for idx, obj in enumerate(DATASET):
                e.process_object(obj.object_type, obj.anonymize() or obj)
                if (
                    isinstance(obj, (Release, Revision))
                    and sensitive_output_path is not None
                ):
                    with open(f"{tmp_dup_dir}/{idx}.csv", "w") as tmp_csv:
                        writer = csv.writer(tmp_csv)
                        if obj.author is not None:
                            _add_person(writer, obj.author)
                        if isinstance(obj, Revision):
                            _add_person(writer, obj.committer)

    if sensitive_output_path is not None:
        for dup_file in tmp_dup_dir.iterdir():
            subprocess.Popen(
                # fmt: off
                [
                    "sort",
                    "-t", ",",
                    "-k", "2",
                    "-u",
                    "-o",
                    tmp_dedup_dir / dup_file.name,
                    dup_file,
                ],
                # fmt: on
                env={**os.environ, "LC_ALL": "C", "LC_COLLATE": "C", "LANG": "C"},
            )

        process_fullnames(
            sensitive_output_path / "orc/person/fullnames.orc", tmp_dedup_dir
        )
        shutil.rmtree(tmp_sensitive_dir)

    if args.compress:
        if (output_path / "compressed").exists():
            shutil.rmtree(output_path / "compressed")
        if (
            sensitive_output_path is not None
            and (sensitive_output_path / "compressed").exists()
        ):
            shutil.rmtree(sensitive_output_path / "compressed")
        sensitive_out_dir = (
            sensitive_output_path / "compressed"
            if sensitive_output_path is not None
            else None
        )
        compress(
            graph_name="example",
            in_dir=output_path / "orc",
            out_dir=output_path / "compressed",
            sensitive_in_dir=sensitive_output_path,
            sensitive_out_dir=sensitive_out_dir,
            check_flavor="example",
            conf={"profile": args.profile},
        )


if __name__ == "__main__":
    main()



# Copyright (C) 2022-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import logging
from pathlib import Path
import subprocess
from typing import TYPE_CHECKING, List

from swh.core.s3.downloader import S3Downloader

if TYPE_CHECKING:
    from types_boto3_s3.service_resource import ObjectSummary


logger = logging.getLogger(__name__)


class GraphDownloader(S3Downloader):
    """Utility class to download a compressed Software Heritage graph dataset
    from S3 implementing a download resumption feature in case some files fail to
    be downloaded (when connection errors happen for instance).

    Example of use::

        from swh.graph.download import GraphDownloader

        # download "2025-05-18-popular-1k" graph dataset into a sub-directory of the
        # current working directory named "2025-05-18-popular-1k"

        graph_downloader = GraphDownloader(
            local_path="2025-05-18-popular-1k",
            s3_url="s3://softareheritage/graph/2025-05-18-popular-1k/compressed/"
        )

        while not graph_downloader.download():
            continue

    """

    def filter_objects(self, objects: List[ObjectSummary]) -> List[ObjectSummary]:
        # meta/compression.json file must be downloaded after all other files
        return [obj for obj in objects if not obj.key.endswith("meta/compression.json")]

    def can_download_file(self, relative_path: str, local_file_path: Path) -> bool:
        # do not download again a file compressed with zstd if it was locally uncompressed
        return (
            not relative_path.endswith(".bin.zst")
            or not Path(str(local_file_path)[:-4]).exists()
        )

    def post_download_file(self, relative_path: str, local_file_path: Path) -> None:
        if (
            relative_path.endswith(".bin.zst")
            and not Path(str(local_file_path)[:-4]).exists()
        ):
            # The file was compressed with zstd before uploading it to S3, we need it
            # to be decompressed locally
            subprocess.check_call(["unzstd", "-d", "-q", "--rm", local_file_path])

    def post_downloads(self) -> None:
        if not self.s3_url.startswith(
            tuple(
                (
                    f"s3://softwareheritage/graph/{year}-"
                    for year in (2018, 2019, 2020, 2021)
                )
            )
        ):
            # skip metadata download for old graphs, they did not have that file yet
            objects = list(self.bucket.objects.filter(Prefix=self.prefix))
            for obj in objects:
                if obj.key.endswith("meta/compression.json"):
                    # Write it last, to act as a stamp
                    self._download_file(obj.key)
                    break
            else:
                raise ValueError(
                    "did not see meta/compression.json in directory listing"
                )



# Copyright (C) 2019-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from pathlib import Path
import shutil
from typing import Dict

from click.testing import CliRunner
import pytest
import yaml

from swh.graph.cli import graph_cli_group
from swh.graph.example_dataset import DATASET_DIR


@pytest.fixture
def cli_runner():
    return CliRunner()


def read_properties(properties_fname) -> Dict[str, str]:
    """read a Java .properties file"""
    with open(properties_fname) as f:
        keyvalues = (
            line.split("=", maxsplit=1)
            for line in f
            if not line.strip().startswith("#")
        )
        return dict((k.strip(), v.strip()) for (k, v) in keyvalues)


def test_pipeline(cli_runner, tmp_path):
    """run full compression pipeline"""
    # bare bone configuration, to allow testing the compression pipeline
    # with minimum RAM requirements on trivial graphs
    config = {
        "profile": "debug",
        "graph": {
            "compress": {
                "batch_size": 1000,
                "rust_executable_dir": "./target/debug/",
                "check_flavor": "example",
            }
        },
    }

    config_path = Path(tmp_path, "config.yml")
    config_path.write_text(yaml.dump(config))

    result = cli_runner.invoke(
        graph_cli_group,
        [
            "--config-file",
            config_path,
            "compress",
            "--input-dataset",
            DATASET_DIR / "orc",
            "--output-directory",
            tmp_path,
            "--sensitive-output-directory",
            tmp_path,
            "--graph-name",
            "example",
        ],
    )
    assert result.exit_code == 0, result
    properties = read_properties(tmp_path / "example.properties")

    assert int(properties["nodes"]) == 24
    assert int(properties["arcs"]) == 28


@pytest.mark.parametrize("option", ["none", "--ef", "--force"])
def test_reindex(cli_runner, mocker, tmp_path, option):
    config = {
        "graph": {
            "compress": {
                "rust_executable_dir": "./target/debug/",
            }
        }
    }

    trailing = []
    if option != "none":
        trailing.append(option)

    config_path = Path(tmp_path, "config.yml")
    config_path.write_text(yaml.dump(config))

    shutil.copytree(DATASET_DIR / "compressed", tmp_path, dirs_exist_ok=True)

    result = cli_runner.invoke(
        graph_cli_group,
        [
            "--config-file",
            config_path,
            "--profile",
            "debug",
            "reindex",
            f"{tmp_path}/example",
            *trailing,
        ],
    )
    assert result.exit_code == 0, result


@pytest.mark.parametrize("exit_code", [0, 1])
def test_luigi(cli_runner, mocker, tmp_path, exit_code):
    """calls Luigi with the given configuration"""
    # bare bone configuration, to allow testing the compression pipeline
    # with minimum RAM requirements on trivial graphs

    subprocess_run = mocker.patch("subprocess.run")
    subprocess_run.return_value.returncode = exit_code

    result = cli_runner.invoke(
        graph_cli_group,
        [
            "luigi",
            "--base-directory",
            f"{tmp_path}/base_dir",
            "--dataset-name",
            "2022-12-07",
            "--",
            "foo",
            "bar",
            "--baz",
            "qux",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == exit_code, result

    luigi_config_path = subprocess_run.mock_calls[0][2]["env"]["LUIGI_CONFIG_PATH"]
    subprocess_run.assert_called_once_with(
        [
            "luigi",
            "--module",
            "swh.export.luigi",
            "--module",
            "swh.graph.luigi",
            "foo",
            "bar",
            "--baz",
            "qux",
        ],
        env={"LUIGI_CONFIG_PATH": luigi_config_path, **os.environ},
    )


def test_download_graph_ok(cli_runner, s3_graph_dataset_name, tmp_path, mocked_aws):
    result = cli_runner.invoke(
        graph_cli_group,
        [
            "download",
            "--name",
            s3_graph_dataset_name,
            str(tmp_path),
        ],
    )
    assert result.exit_code == 0, result.output
    assert (tmp_path / "example.graph").exists()
    assert (tmp_path / "meta/compression.json").exists()

    assert (
        f"Graph dataset {s3_graph_dataset_name} successfully downloaded to path {tmp_path}."
    ) in result.output


def test_download_graph_resumption(
    cli_runner, s3_graph_dataset_name, mocker, tmp_path, mocked_aws
):
    mocker.patch("swh.graph.download.GraphDownloader.download").side_effect = [
        False,
        True,
    ]
    result = cli_runner.invoke(
        graph_cli_group,
        [
            "download",
            "--name",
            s3_graph_dataset_name,
            str(tmp_path),
        ],
    )
    assert result.exit_code == 0, result.output

    assert (
        f"Graph dataset {s3_graph_dataset_name} successfully downloaded to path {tmp_path}."
    ) in result.output



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from pathlib import Path

import boto3
import pytest


def add_example_dataset_to_s3_bucket(
    datasets_path: Path, bucket: str, prefix: str, name: str
):
    s3 = boto3.client("s3", region_name="us-east-1")
    s3.create_bucket(Bucket=bucket)
    for path in datasets_path.rglob("**/*"):
        if path.is_file():
            relative_path = path.relative_to(datasets_path)
            key = os.path.join(prefix, name, "compressed", str(relative_path))
            s3.upload_file(
                Filename=str(path),
                Bucket=bucket,
                Key=key,
                ExtraArgs={
                    "ACL": "public-read",
                },
            )
    s3.put_object(
        ACL="public-read",
        Body=b"{}",
        Bucket=bucket,
        Key=os.path.join(prefix, name, "compressed/meta/compression.json"),
    )


@pytest.fixture
def s3_bucket_name():
    return "softwareheritage"


@pytest.fixture
def s3_graph_dataset_path_prefix():
    return "graph"


@pytest.fixture
def s3_graph_dataset_name():
    return "example"


@pytest.fixture
def s3_graph_dataset_url(
    s3_bucket_name, s3_graph_dataset_path_prefix, s3_graph_dataset_name
):
    return (
        f"s3://{s3_bucket_name}/{s3_graph_dataset_path_prefix}/"
        f"{s3_graph_dataset_name}/compressed/"
    )


@pytest.fixture
def graph_example_dataset_path():
    import swh

    return Path(swh.graph.__file__).parent / "example_dataset" / "compressed"


@pytest.fixture
def mocked_aws(
    mocked_aws,
    graph_example_dataset_path,
    s3_bucket_name,
    s3_graph_dataset_path_prefix,
    s3_graph_dataset_name,
):
    add_example_dataset_to_s3_bucket(
        graph_example_dataset_path,
        s3_bucket_name,
        s3_graph_dataset_path_prefix,
        s3_graph_dataset_name,
    )
